# Tabular ML models and benchmark

This repository contains code for applying tabular ML methods 
and benchmarking them on our meta-train and meta-test benchmarks. 
The tabular ML methods are contained in `tab_models`. 
The benchmarking code, which uses `tab_models` is contained in `tab_bench`.

## Installation

We use python (tested with python 3.9 and 3.10).
The required libraries can be installed as follows:

```commandline
# install faiss-gpu or faiss-cpu (via conda) if you want to use TabR
pip3 install wheel  # for simpler installation
pip3 install torch  # use your preferred installation variant
# pip3 install swig  # might be helpful in case building pyrfr wheels fails
pip3 install numpy dill scikit-learn xgboost catboost lightgbm torchmetrics pyyaml hyperopt smac pytorch-lightning mkdocs-material 'mkdocstrings[python]' msgpack ConfigSpace kditransform skorch msgpack_numpy
# only if you want to use tab_bench and high-level scripts, not needed for tab_models
sudo apt install rar  # or 7-zip or unrar, needed for unpacking some of the UCI datasets for meta-train
pip3 install matplotlib fire ray pynvml psutil openml requests patool openpyxl xlrd tueplots seaborn adjustText skorch distinctipy cir_model venn_abers imodelsx autorank
# for docs
# pip3 install mkdocs-material 'mkdocstrings[python]' black
pip3 install sphinx myst_parser
# for test
pip3 install pytest pytest-cov
```


## Basic setup

To configure the data path for `tab_bench`, configure the environment variable `TAB_BENCH_DATA_BASE_FOLDER` by
```commandline
export TAB_BENCH_DATA_BASE_FOLDER=<path>
```
Alternatively, copy `custom_paths.py.default` to `custom_paths.py` and adjust the path inside the file to 
wherever you want your data to be stored if you run the benchmarking code 
(`custom_paths.py` will be ignored by git so everyone can have their own configuration).


## Using the ML models

Most of our machine learning models are directly available via scikit-learn interfaces.
For example, you can use RealMLP-TD for classification as follows:

```python
from tab_models.sklearn.sklearn_interfaces import RealMLP_TD_Classifier

model = RealMLP_TD_Classifier()
model.fit(X_train, y_train)
model.predict(X_test)
```
The code above will automatically select a GPU if available, 
try to detect categorical columns in dataframes, 
preprocess numerical variables and regression targets (no standardization required),
and use a training-validation split for early stopping. 
All of this (and much more) can be configured through the constructor 
and the parameters of the fit() method. 
For example, it is possible to do bagging 
(ensembling of models on 5-fold cross-validation)
simply by passing `n_cv=5` to the constructor. 
Here is an example for some of the parameters that can be set explicitly:
```python
from tab_models.sklearn.sklearn_interfaces import RealMLP_TD_Classifier

model = RealMLP_TD_Classifier(device='cpu', random_state=0, n_cv=1, n_refit=0,
                              verbosity=2, val_metric_name='cross_entropy', 
                              n_epochs=256, batch_size=256, hidden_sizes=[256]*3, 
                              lr=0.04, use_ls=False)
model.fit(X_train, y_train, val_idxs=val_idxs, cat_features=cat_features)
model.predict_proba(X_test)
```

## Running the benchmark

To download all datasets for the meta-train and meta-test benchmarks, run (with your desired OpenML cache directory)
```commandline
python3 download_data.py openml_cache_dir
```
To run methods on the benchmarks, there are two options:

### With slurm

Run the following command (replacing some of the parameters with your own values) on the login node:
```commandline
python3 ray_slurm_launch.py --exp_name=my_exp_name --num_nodes=num_nodes --queue="queue_name" --time=24:00:00 --mail_user="my@address.edu" --log_folder=log_folder --command="python3 -u run_slurm.py"
```
This will submit a job to the configured queue that will run `run_slurm.py` and create logfiles.
Your experiments then have to be configured in `run_slurm.py`, see below.
Multi-node is supported: `ray` will start instances on each node
and our benchmarking code will schedule the individual experiments on the nodes.

### Without slurm

Run the file with the corresponding experiments directly. 
For example, many of our experiment configurations 
can be found in `run_final_experiments.py`. 
One possible way to run the experiments detached from the shell with log-files is
````commandline
systemd-run --scope --user python3 -u run_final_experiments.py > ./out.log 2> ./err.log &
````

### Time measurements

For time measurements, simply run `run_time_measurements.py` (with or without slurm).
Results can be printed using `print_runtimes.py`.

## Evaluating the benchmark results

Aggregated algorithm results can be printed using 
````commandline
python3 run_evaluation.py meta-train-class
````
where meta-train-class can be replaced by the name of any other task collection,
or a single dataset such as `openml-class/Higgs`.
This script also has many more command line options, see the python file.
For example, one can print only those methods with a certain tag 
using the `--tag` option,
print results on individual datasets, for different metrics, etc.

## Creating plots and tables

Plots and tables can be created using
````commandline
python3 create_plots_and_tables.py
````

## Single-task experiments

You can also run a configuration on a single data set,
without saving the results, by adjusting and running `single_task.py`.

## Other utilities

- Use `analyze_tasks.py` to print some dataset statistics.
- You can rename a method using `python3 rename_alg.py old_name new_name`.
- We used some code in `meta_hyperopt.py` to optimize the default parameters for GBDTs.
- The code in `estimate_resource_params.py` has been used to get more precise estimates 
for RAM usage etc. for running methods on the benchmark.
- `print_complete_results.py` can be used to check which methods have results available on all splits for all tasks in a given collection.

## Using Sphinx Documentation
Go to the repo root dir and simply run
`sphinx-build -M html docs/source/ docs/build/`

## Acknowledgements
Code from other repositories is acknowledged as well as possible in code comments. 
Especially, we used code from https://github.com/yandex-research/rtdl 
and sub-packages (Apache 2.0 license),
code from https://github.com/catboost/benchmarks/
(Apache 2.0 license), 
and https://docs.ray.io/en/latest/cluster/vms/user-guides/community/slurm.html 
(Apache 2.0 license).