from typing import List, Dict, Any

from sklearn.base import BaseEstimator

import custom_paths
from tab_models.alg_interfaces.resource_computation import UniformSampler, FeatureSpec, get_resource_features, process_resource_features, \
    calibrate_resources_new, calibrate_resources_new_2
from tab_bench.data.paths import Paths
from tab_bench.data.tasks import TaskDescription
from tab_models import utils
from tab_models.data import DictDataset
from tab_models.sklearn.sklearn_interfaces import CatBoost_TD_Classifier, XGB_TD_Classifier, LGBM_TD_Classifier


def get_param_grid(grids_1d: Dict[str, List[Any]]) -> List[Dict[str, Any]]:
    configs = [dict()]
    for key, values in grids_1d.items():
        configs = [utils.update_dict(c, {key: val}) for val in values for c in configs]
    return configs


def estimate_params(paths: Paths, exp_name: str, coll_name: str, estimator: BaseEstimator, is_lgbm: bool = False,
                    rerun: bool = False):
    if is_lgbm:
        # use num_leaves instead of max_depth
        learner_space = dict(
            n_estimators=UniformSampler(2, 2, log=True, is_int=True),
            n_threads=UniformSampler(4, 4, log=True, is_int=True),
            num_leaves=UniformSampler(10, 100, log=True, is_int=True),
        )
        time_feature_spec = FeatureSpec.concat('', 'ds_size_gb', 'ds_prep_size_gb', 'ds_onehot_size_gb',
                                               FeatureSpec.product('n_cv_refit', 'n_splits',
                                                                   ['', 'log_num_leaves', 'num_leaves'],
                                                                   'n_estimators', '1/n_threads',
                                                                   FeatureSpec.powerset_products('n_features',
                                                                                                 'n_samples',
                                                                                                 'n_tree_repeats')))
        ram_feature_spec = FeatureSpec.concat('', 'ds_size_gb', 'ds_prep_size_gb', 'ds_onehot_size_gb',
                                              FeatureSpec.product(['', 'log_num_leaves', 'num_leaves'],
                                                                  FeatureSpec.powerset_products('n_features',
                                                                                                'n_samples',
                                                                                                'n_tree_repeats')))
    else:
        learner_space = dict(
            n_estimators=UniformSampler(2, 2, log=True, is_int=True),
            n_threads=UniformSampler(4, 4, log=True, is_int=True),
            max_depth=UniformSampler(3, 10, is_int=True),
        )
        time_feature_spec = FeatureSpec.concat('', 'ds_size_gb', 'ds_prep_size_gb', 'ds_onehot_size_gb',
                                               FeatureSpec.product('n_cv_refit', 'n_splits',
                                                                   ['', 'max_depth', '2_power_maxdepth'],
                                                                   'n_estimators', '1/n_threads',
                                                                   FeatureSpec.powerset_products('n_features',
                                                                                                 'n_samples',
                                                                                                 'n_tree_repeats')))
        ram_feature_spec = FeatureSpec.concat('', 'ds_size_gb', 'ds_prep_size_gb', 'ds_onehot_size_gb',
                                              FeatureSpec.product(['', 'max_depth', '2_power_maxdepth'],
                                                                  FeatureSpec.powerset_products('n_features',
                                                                                                'n_samples',
                                                                                                'n_tree_repeats')))

    coefs = calibrate_resources_new(exp_name, paths=paths, learner_space=learner_space,
                                    coll_name=coll_name,
                                    time_feature_spec=time_feature_spec,
                                    ram_feature_spec=ram_feature_spec, sklearn_learner=estimator, n_combinations=300,
                                    rerun=rerun)
    print(f'time_params={coefs["time_s"]}')
    print(f'cpu_ram_params={coefs["ram_gb"]}')

    ram_params = coefs['ram_gb']
    time_params = coefs['time_s']

    print(f'Analyzing dionis:')
    task_info = TaskDescription('openml-class', 'dionis').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'madelon').load_info(paths)
    ds = DictDataset(tensors=None, tensor_infos=task_info.tensor_infos, device='cpu', n_samples=task_info.n_samples)
    config = dict(n_estimators=1000, n_threads=4, max_depth=6, num_leaves=31)
    raw_features = get_resource_features(config, ds, n_cv=1, n_refit=0, n_splits=1)
    ram_features = process_resource_features(raw_features, ram_feature_spec)
    ram_gb = sum([ram_features[key] * ram_params[key] for key in ram_params])
    time_features = process_resource_features(raw_features, time_feature_spec)
    time_s = sum([time_features[key] * time_params[key] for key in time_params])
    print(f'{ram_gb=}, {time_s=}')


def estimate_params_new(paths: Paths, exp_name: str, coll_name: str, estimator: BaseEstimator,
                        hparam_grid: List[Dict[str, Any]], short_name: str, is_lgbm: bool = False, rerun: bool = False):
    if is_lgbm:
        # use num_leaves instead of max_depth
        time_feature_spec = FeatureSpec.concat('', 'ds_size_gb', 'ds_prep_size_gb', 'ds_onehot_size_gb',
                                               FeatureSpec.product('n_cv_refit', 'n_splits',
                                                                   ['', 'log_num_leaves', 'num_leaves'],
                                                                   'n_estimators', '1/n_threads',
                                                                   FeatureSpec.powerset_products('n_features',
                                                                                                 'n_samples',
                                                                                                 'n_tree_repeats')))
        ram_feature_spec = FeatureSpec.concat('', 'ds_size_gb', 'ds_prep_size_gb', 'ds_onehot_size_gb',
                                              FeatureSpec.product(['', 'log_num_leaves', 'num_leaves'],
                                                                  FeatureSpec.powerset_products('n_features',
                                                                                                'n_samples',
                                                                                                'n_tree_repeats')))
    else:
        time_feature_spec = FeatureSpec.concat('', 'ds_size_gb', 'ds_prep_size_gb', 'ds_onehot_size_gb',
                                               FeatureSpec.product('n_cv_refit', 'n_splits',
                                                                   ['', 'max_depth', '2_power_maxdepth'],
                                                                   'n_estimators', '1/n_threads',
                                                                   FeatureSpec.powerset_products('n_features',
                                                                                                 'n_samples',
                                                                                                 'n_tree_repeats')))
        ram_feature_spec = FeatureSpec.concat('', 'ds_size_gb', 'ds_prep_size_gb', 'ds_onehot_size_gb',
                                              FeatureSpec.product(['', 'max_depth', '2_power_maxdepth'],
                                                                  FeatureSpec.powerset_products('n_features',
                                                                                                'n_samples',
                                                                                                'n_tree_repeats')))

    coefs = calibrate_resources_new_2(exp_name, paths=paths, hparam_grid=hparam_grid,
                                      coll_name=coll_name,
                                      time_feature_spec=time_feature_spec,
                                      ram_feature_spec=ram_feature_spec, sklearn_learner=estimator,
                                      rerun=rerun)
    print(f'{short_name}_time={coefs["time_s"]}')
    print(f'{short_name}_ram={coefs["ram_gb"]}')

    ram_params = coefs['ram_gb']
    time_params = coefs['time_s']

    print(f'Analyzing dionis:')
    task_info = TaskDescription('openml-class', 'dionis').load_info(paths)
    # task_info = TaskDescription('uci-bin-class', 'madelon').load_info(paths)
    ds = DictDataset(tensors=None, tensor_infos=task_info.tensor_infos, device='cpu', n_samples=task_info.n_samples)
    config = dict(n_estimators=1000, n_threads=4, max_depth=6, num_leaves=31)
    raw_features = get_resource_features(config, ds, n_cv=1, n_refit=0, n_splits=1)
    ram_features = process_resource_features(raw_features, ram_feature_spec)
    ram_gb = sum([ram_features[key] * ram_params[key] for key in ram_params])
    time_features = process_resource_features(raw_features, time_feature_spec)
    time_s = sum([time_features[key] * time_params[key] for key in time_params])
    print(f'{ram_gb=}, {time_s=}')


if __name__ == '__main__':
    print(get_param_grid(dict(n_estimators=[2], max_depth=[4, 6, 7, 9])))
    paths = Paths(base_folder=custom_paths.get_base_folder())
    # estimate_catboost_params(paths)
    # estimate_params(paths, 'CB-class-7', 'meta-test-class', CatBoostTDClassifier(verbosity=2))
    # # estimate_params(paths, 'CB-reg-7', 'meta-test-reg', CatBoostTDClassifier(verbosity=2))
    # estimate_params(paths, 'XGB-class-2', 'meta-test-class',
    #                 XGBTDClassifier(verbosity=2, subsample=1.0, colsample_bytree=1.0, colsample_bylevel=1.0))
    # estimate_params(paths, 'LGBM-class-3', 'meta-test-class',
    #                 LGBMTDClassifier(subsample=1.0), is_lgbm=True)
    estimate_params_new(paths, 'CB-class-11', 'meta-test-class',
                        CatBoost_TD_Classifier(subsample=1.0),
                        hparam_grid=get_param_grid(dict(n_estimators=[2], n_threads=[4], max_depth=[4, 6, 7, 9])),
                        short_name='cb_class')
    # estimate_params(paths, 'CB-reg-7', 'meta-test-reg', CatBoostTDClassifier(verbosity=2))
    estimate_params_new(paths, 'XGB-class-3', 'meta-test-class',
                        XGB_TD_Classifier(subsample=1.0, colsample_bytree=1.0, colsample_bylevel=1.0),
                        hparam_grid=get_param_grid(dict(n_estimators=[2], n_threads=[4], max_depth=[4, 6, 8, 11])),
                        short_name='xgb_class')
    estimate_params_new(paths, 'LGBM-class-4', 'meta-test-class',
                        LGBM_TD_Classifier(subsample=1.0, colsample_bytree=1.0),
                        hparam_grid=get_param_grid(dict(n_estimators=[2], n_threads=[4],
                                                        num_leaves=[31, 100, 300, 1000])),
                        short_name='lgbm_class',
                        is_lgbm=True)
    pass
