import time
from typing import Optional

import numpy as np

import fire

from tab_bench.data.common import SplitType
from tab_bench.data.paths import Paths
from tab_bench.data.tasks import TaskDescription, TaskCollection
from tab_bench.eval.analysis import get_opt_groups
from tab_bench.eval.evaluation import MultiResultsTable, DefaultEvalModeSelector, MeanTableAnalyzer, alg_results_str, \
    alg_comparison_str, WinsTableAnalyzer, RankTableAnalyzer, NormalizedLossTableAnalyzer, \
    GreedyAlgSelectionTableAnalyzer


def show_eval(coll_name: str = 'meta-train-class', n_cv: int = 1, show_alg_groups: bool = True,
              val_metric_name: str = None, metric_name: str = None, split_type: str = SplitType.RANDOM,
              use_task_weighting: Optional[bool] = None, shift_eps: float = 0.01,
              data_path: Optional[str] = None, alg_name: Optional[str] = None,
              alg_name_2: Optional[str] = None, tag: Optional[str] = None, max_n_splits: Optional[int] = None):
    # preference regarding is_cv and ensembling?
    # optionally whether default splits should be used or not?
    # tags (connect by and or or?)
    print('start show eval')
    paths = Paths(data_path) if data_path is not None else Paths.from_env_variables()
    start_time = time.time()
    if '/' in coll_name:
        # use a single task
        parts = coll_name.split('/')
        if len(parts) != 2:
            print(f'Too many / in coll_name {coll_name}')
            return
        task_collection = TaskCollection(coll_name, [TaskDescription(*parts)])
    else:
        task_collection = TaskCollection.from_name(coll_name, paths)
    print('load table')
    # table = MultiResultsTable.load_summaries(task_collection, n_cv=n_cv, paths=paths)
    if tag is None:
        alg_filter = None
    else:
        show_tags = tag.split(',') if isinstance(tag, str) else list(
            tag)  # commas are converted to tuples in the command line, apparently
        alg_filter = lambda an, tags, config: np.any([show_tag in tags for show_tag in show_tags])
    table = MultiResultsTable.load(task_collection, n_cv=n_cv, paths=paths,
                                   split_type=split_type, alg_filter=alg_filter, max_n_splits=max_n_splits)
    print('process table')
    # alg_group_dict = {'all algs': (lambda an, tags, config: True)} if show_alg_groups else None
    task_type_name = 'class' if 'class' in coll_name else 'reg'
    opt_groups = get_opt_groups(task_type_name)
    alg_group_dict = {'BestModel': (lambda an, tags, config: not an.startswith('Ensemble')), **{
        f'BestModel{group_name}': (lambda an, tags, config, ans=alg_names: an in ans)
        for group_name, alg_names in opt_groups.items()
    }}
    if not show_alg_groups:
        alg_group_dict = None
    if alg_name is not None and alg_name_2 is not None:
        alg_group_dict['selected algs'] = (lambda an, tags, config, grp=[alg_name, alg_name_2]:
                                           np.any([g.startswith(an) for g in grp]))

    val_test_groups = {f'HPO-on-BestModel-TD-{task_type_name}': {f'{family}-TD-{task_type_name}': f'{family}-HPO'
                                               for family in ['XGB', 'LGBM', 'CatBoost', 'MLP']}
                       for task_type_name in ['class', 'reg']}

    test_table = table.get_test_results_table(DefaultEvalModeSelector(), alg_group_dict=alg_group_dict,
                                              test_metric_name=metric_name, val_metric_name=val_metric_name,
                                              val_test_groups=val_test_groups)
    val_table_single = table.get_test_results_table(DefaultEvalModeSelector(), alg_group_dict=dict(),
                                              test_metric_name=metric_name, val_metric_name=val_metric_name,
                                              val_test_groups=val_test_groups, use_validation_errors=True)
    test_table_single = table.get_test_results_table(DefaultEvalModeSelector(), alg_group_dict=dict(),
                                                    test_metric_name=metric_name, val_metric_name=val_metric_name,
                                                    val_test_groups=val_test_groups, use_validation_errors=False)

    if use_task_weighting is None:
        use_task_weighting = coll_name.startswith('meta-train') or coll_name.startswith('uci')
    separate_task_names = ['facebook_comment_volume', 'facebook_live_sellers_thailand_shares']
    print(f'Greedy algorithm selection cumulative best log shifted geometric mean (err+{shift_eps:g}) test error:')
    analyzer = GreedyAlgSelectionTableAnalyzer(use_weighting=use_task_weighting,
                                               separate_task_names=separate_task_names,
                                               f=lambda x: np.log(x + shift_eps))
    analyzer.print_analysis(test_table_single, val_table_single)
    print()
    print('Win fraction:')
    analyzer = WinsTableAnalyzer(use_weighting=use_task_weighting, separate_task_names=separate_task_names)
    analyzer.print_analysis(test_table)
    print()
    print('Arithmetic mean rank:')
    analyzer = RankTableAnalyzer(use_weighting=use_task_weighting, separate_task_names=separate_task_names)
    analyzer.print_analysis(test_table)
    print()
    print('Arithmetic mean normalized test error:')
    analyzer = NormalizedLossTableAnalyzer(use_weighting=use_task_weighting, separate_task_names=separate_task_names)
    analyzer.print_analysis(test_table)
    print()
    print('Arithmetic mean test error:')
    analyzer = MeanTableAnalyzer(use_weighting=use_task_weighting, separate_task_names=separate_task_names)
    analyzer.print_analysis(test_table)
    print()
    print(f'Log shifted geometric mean (err+{shift_eps:g}) test error:')
    analyzer = MeanTableAnalyzer(f=lambda x: np.log(x + shift_eps), use_weighting=use_task_weighting,
                                 separate_task_names=separate_task_names)
    analyzer.print_analysis(test_table)
    print()
    print(f'Shifted geometric mean (err+{shift_eps:g}) test error:')
    analyzer = MeanTableAnalyzer(f=lambda x: np.log(x + shift_eps), use_weighting=use_task_weighting,
                                 separate_task_names=separate_task_names,
                                 post_f=lambda x: np.exp(x))
    analyzer.print_analysis(test_table)
    print()
    # print('Mean modlog test error:')  # todo: name modlog is suboptimal, people could associate mod with modulo
    # analyzer = MeanTableAnalyzer(f=lambda x: np.log(x + 1e-3) - np.log(1e-3), use_weighting=use_task_weighting)
    # analyzer.print_analysis(test_table)
    if alg_name is not None:
        if alg_name_2 is None:
            print(f'Errors for alg {alg_name}:')
            print(alg_results_str(test_table, alg_name))
        else:
            print(f'Comparison: {alg_name} vs. {alg_name_2}')
            print(alg_comparison_str(test_table, [alg_name, alg_name_2]))
    print(f'Time for printing: {time.time() - start_time:g} s')


if __name__ == '__main__':
    fire.Fire(show_eval)
