from pathlib import Path
from typing import List, Optional, Tuple, Dict, Any

import numpy as np
import sklearn
import torch
import torch.nn as nn
from cir_model import CenteredIsotonicRegression
from sklearn.base import BaseEstimator
from sklearn.calibration import CalibratedClassifierCV
from venn_abers import VennAbers

from tab_models.alg_interfaces.alg_interfaces import AlgInterface
from tab_models.alg_interfaces.base import SplitIdxs, InterfaceResources, RequiredResources
from tab_models.data import DictDataset
from tab_models.training.logging import Logger


class ApplyToLogitsCalibrator(BaseEstimator):
    def __init__(self, calib: BaseEstimator):
        self.calib = calib

    def fit(self, X, y):
        self.calib_ = sklearn.base.clone(self.calib)
        self.calib_.fit(np.log(X + 1e-10), y)
        return self

    def predict_proba(self, X):
        return self.calib_.predict_proba(np.log(X + 1e-10))


class TemperatureScalingCalibrator(BaseEstimator):
    # following https://github.com/gpleiss/temperature_scaling/blob/master/temperature_scaling.py
    def _get_logits(self, X):
        X = X + 1e-10
        X /= np.sum(X, axis=-1, keepdims=True)
        return torch.as_tensor(np.log(X), dtype=torch.float32)

    def fit(self, X, y):
        logits = self._get_logits(X)
        labels = torch.as_tensor(y)
        # todo: do we need to prevent temperature from going negative? Maybe use inverse temperature instead?
        self.temperature_ = nn.Parameter(torch.ones(1) * 1.5)
        criterion = nn.CrossEntropyLoss()

        optimizer = torch.optim.LBFGS([self.temperature_], lr=0.01, max_iter=50)

        def eval():
            optimizer.zero_grad()
            y_pred = logits / self.temperature_[:, None]
            loss = criterion(y_pred, labels)
            loss.backward()
            return loss

        for i in range(50):
            optimizer.step(eval)

        # print(f'Optimal temperature: {self.temperature_.item():g}')
        return self

    def predict_proba(self, X):
        logits = self._get_logits(X)
        with torch.no_grad():
            y_pred = logits / self.temperature_[:, None]
            return torch.softmax(y_pred, dim=-1).detach().numpy()


class InverseTemperatureScalingCalibrator(BaseEstimator):
    # following https://github.com/gpleiss/temperature_scaling/blob/master/temperature_scaling.py
    def _get_logits(self, X):
        X = X + 1e-10
        X /= np.sum(X, axis=-1, keepdims=True)
        return torch.as_tensor(np.log(X), dtype=torch.float32)

    def fit(self, X, y):
        logits = self._get_logits(X)
        labels = torch.as_tensor(y)
        self.inv_temperature_ = nn.Parameter(torch.ones(1) / 1.5)
        criterion = nn.CrossEntropyLoss()

        optimizer = torch.optim.LBFGS([self.inv_temperature_], lr=0.01, max_iter=50)

        def eval():
            optimizer.zero_grad()
            y_pred = logits * self.inv_temperature_[:, None]
            loss = criterion(y_pred, labels)
            loss.backward()
            return loss

        for i in range(50):
            optimizer.step(eval)

        print(f'Optimal temperature: {(1./self.inv_temperature_).item():g}')
        return self

    def predict_proba(self, X):
        logits = self._get_logits(X)
        with torch.no_grad():
            y_pred = logits * self.inv_temperature_[:, None]
            return torch.softmax(y_pred, dim=-1).detach().numpy()


class CenteredIsotonicRegressionCalibrator(BaseEstimator):
    def fit(self, X, y):
        assert not np.any(np.isnan(X))
        # print(f'{np.unique(y)=}')
        # print(f'{np.unique(X)=}')
        # we have to use float64 since with float32 it can happen that somewhere internally
        # a rounding error occurs (?) and interp1d thinks that it is asked to extrapolate at the boundary value,
        # which causes it to output nan values,
        # which are seen as a separate possible value but (y==value) is empty because nan==nan is false,
        # so an error occurs because the method attempts to take an average of an empty set
        X = X.astype(np.float64)
        y = y.astype(np.float64)
        self.cal_ = CenteredIsotonicRegression()
        self.cal_.fit(X[:, 1], y)
        self.min_ = np.min(X[:, 1])
        self.max_ = np.max(X[:, 1])
        return self

    def predict_proba(self, X):
        # have to clip since CenteredIsotonicRegression refuses to extrapolate (?)
        X = X.astype(np.float64)
        X = np.clip(X, self.min_, self.max_)
        pred_probs = self.cal_.transform(X[:, 1])
        pred_probs = np.clip(pred_probs, 0.0, 1.0)
        return np.stack([1. - pred_probs, pred_probs], axis=-1)


class BinaryVennAbersCalibrator(BaseEstimator):
    def fit(self, X, y):
        self.va_ = VennAbers()
        self.va_.fit(X, y)
        return self

    def predict_proba(self, X):
        return self.va_.predict_proba(X)[0]


class MulticlassOneVsOneCalibrator(BaseEstimator):
    def __init__(self, binary_calibrator):
        self.binary_calibrator = binary_calibrator

    def fit(self, X, y):
        self.n_classes_ = X.shape[-1]
        self.classes_ = list(range(self.n_classes_))
        self.bin_calibs_ = []

        if self.n_classes_ == 2:
            # binary classification
            bin_calib = sklearn.base.clone(self.binary_calibrator)
            bin_calib.fit(X, y)
            self.bin_calibs_.append(bin_calib)
        else:
            for i in range(self.n_classes_):
                for j in range(i + 1, self.n_classes_):
                    idxs = np.logical_or((y == i), (y == j))
                    # idxs = np.arange(X.shape[0])
                    bin_probs = np.stack([X[idxs, j], X[idxs, i]], axis=-1)
                    bin_probs += 1e-30
                    bin_probs /= np.sum(bin_probs, axis=-1, keepdims=True)
                    print(f'{np.any(np.isnan(bin_probs))=}')
                    bin_labels = (y[idxs] == i).astype(np.int32)
                    bin_calib = sklearn.base.clone(self.binary_calibrator)
                    bin_calib.fit(bin_probs, bin_labels)
                    self.bin_calibs_.append(bin_calib)

        return self

    def predict_proba(self, X):
        if self.n_classes_ == 2:
            # binary classification
            return self.bin_calibs_[0].predict_proba(X)
        else:
            # use PKPD formula in http://proceedings.mlr.press/v60/manokhin17a/manokhin17a.pdf
            pair_probs = [[None] * self.n_classes_ for i in range(self.n_classes_)]
            multi_probs = []
            calib_idx = 0
            for i in range(self.n_classes_):
                for j in range(i + 1, self.n_classes_):
                    bin_probs = np.stack([X[:, j], X[:, i]], axis=-1)
                    bin_probs += 1e-30
                    bin_probs /= np.sum(bin_probs, axis=-1, keepdims=True)
                    pred_probs = self.bin_calibs_[calib_idx].predict_proba(bin_probs)
                    pair_probs[i][j] = pred_probs[:, 1]
                    pair_probs[j][i] = pred_probs[:, 0]
                    if i == 0 and j == 1:
                        print(f'{bin_probs=}')
                        print(f'{pred_probs=}')
                    calib_idx += 1

            for i in range(self.n_classes_):
                sum_inv_probs = sum([1. / (1e-30 + pair_probs[i][j]) for j in range(self.n_classes_) if j != i])
                multi_probs.append(1. / np.clip(sum_inv_probs - (self.n_classes_ - 2), 1e-30, np.Inf))

            multi_probs = np.stack(multi_probs, axis=-1)
            multi_probs = np.clip(multi_probs, a_min=1e-30, a_max=np.Inf)
            multi_probs = multi_probs / np.sum(multi_probs, axis=-1, keepdims=True)
            return multi_probs


class MulticlassOneVsRestCalibrator(BaseEstimator):
    def __init__(self, binary_calibrator):
        self.binary_calibrator = binary_calibrator

    def fit(self, X, y):
        self.n_classes_ = X.shape[-1]
        self.classes_ = list(range(self.n_classes_))
        self.bin_calibs_ = []

        X[np.isnan(X)] = 0.0

        if self.n_classes_ == 2:
            # binary classification
            bin_calib = sklearn.base.clone(self.binary_calibrator)
            bin_calib.fit(X, y)
            self.bin_calibs_.append(bin_calib)
        else:
            for i in range(self.n_classes_):
                pos_probs = X[:, i]
                neg_probs = 1. - pos_probs
                bin_probs = np.stack([neg_probs, pos_probs], axis=-1)
                bin_labels = (y == i).astype(np.int32)
                bin_calib = sklearn.base.clone(self.binary_calibrator)
                bin_calib.fit(bin_probs, bin_labels)
                self.bin_calibs_.append(bin_calib)

        return self

    def predict_proba(self, X):
        X[np.isnan(X)] = 0.0
        if self.n_classes_ == 2:
            # binary classification
            return self.bin_calibs_[0].predict_proba(X)
        else:
            multi_probs = []
            for i in range(self.n_classes_):
                pos_probs = X[:, i]
                neg_probs = 1. - pos_probs
                bin_probs = np.stack([neg_probs, pos_probs], axis=-1)
                pred_probs = self.bin_calibs_[i].predict_proba(bin_probs)
                multi_probs.append(pred_probs[:, 1])
            multi_probs = np.stack(multi_probs, axis=-1)
            multi_probs = np.clip(multi_probs, a_min=1e-30, a_max=np.Inf)
            multi_probs = multi_probs / np.sum(multi_probs, axis=-1, keepdims=True)
            return multi_probs


class IdentityEstimator(BaseEstimator):
    def __init__(self, n_classes: int):
        self.n_classes = n_classes
        self.classes_ = list(range(n_classes))

    def fit(self, X, y):
        self.classes_ = np.unique(y)
        return self

    def predict_proba(self, X):
        return X


def get_calibrator(n_classes: int, **config):
    method = config['calibration_method']
    est = IdentityEstimator(n_classes=n_classes)
    # need to have est.classes_ in CalibratedClassifierCV
    # est.fit(X, y)
    if method == 'platt':
        return CalibratedClassifierCV(est, method='sigmoid', cv='prefit')
    elif method == 'platt-logits':
        return ApplyToLogitsCalibrator(CalibratedClassifierCV(est, method='sigmoid', cv='prefit'))
    elif method == 'isotonic':
        return CalibratedClassifierCV(est, method='isotonic', cv='prefit')
    elif method == 'ivap-naive-ovr':
        return MulticlassOneVsRestCalibrator(BinaryVennAbersCalibrator())
    elif method == 'ivap-naive-ovo':
        return MulticlassOneVsOneCalibrator(BinaryVennAbersCalibrator())
    elif method == 'isotonic-naive-ovr':
        # should be the same as 'isotonic', this is just to check
        return MulticlassOneVsRestCalibrator(
            CalibratedClassifierCV(IdentityEstimator(n_classes=2), method='isotonic', cv='prefit'))
    elif method == 'cir-naive-ovr':
        return MulticlassOneVsRestCalibrator(CenteredIsotonicRegressionCalibrator())
    elif method == 'temp-scaling':
        return TemperatureScalingCalibrator()
    elif method == 'inv-temp-scaling':
        return InverseTemperatureScalingCalibrator()
    else:
        raise ValueError(f'Unknown calibration method "{method}"')


class PostHocCalibrationAlgInterface(AlgInterface):
    def __init__(self, alg_interface: AlgInterface, fit_params: Optional[List[Dict[str, Any]]] = None, **config):
        super().__init__(fit_params=fit_params, **config)
        self.alg_interface = alg_interface
        self.calibrators = []
        self.n_calibs = []

    def fit(self, ds: DictDataset, idxs_list: List[SplitIdxs], interface_resources: InterfaceResources,
            logger: Logger, tmp_folders: List[Optional[Path]], name: str) \
            -> Optional[List[List[List[Tuple[Dict, float]]]]]:
        self.alg_interface.fit(ds, idxs_list, interface_resources, logger, tmp_folders, name)
        y_preds = self.alg_interface.predict(ds)
        n_tt_splits = len(idxs_list)
        n_tv_splits = idxs_list[0].n_trainval_splits
        for tt_split_idx in range(n_tt_splits):
            for tv_split_idx in range(n_tv_splits):
                pass

        for split_idxs in idxs_list:
            for tv_split_idx in range(split_idxs.n_trainval_splits):
                val_idxs = split_idxs.val_idxs[tv_split_idx]
                y = ds.tensors['y'][val_idxs]
                y_pred = y_preds[len(self.calibrators), val_idxs]
                y_pred_probs = torch.softmax(y_pred, dim=-1)

                x_calib = y_pred_probs.detach().cpu().numpy()
                y_calib = y.cpu().numpy()[:, 0]
                calib = get_calibrator(n_classes=ds.tensor_infos['y'].get_cat_size_product(), **self.config)
                calib.fit(x_calib, y_calib)

                # calib_probs = calib.predict_proba(y_pred_probs.detach().cpu().numpy())
                # print(f'{y=}')
                # print(f'{y_pred_probs.detach().cpu().numpy().shape=}')
                # print(f'{y.cpu().numpy()[..., 0].shape=}')
                # print(f'{np.unique(y.cpu().numpy())=}')
                # print(f'{y_pred_probs=}')
                # print(f'{calib_probs=}')
                # print(f'{np.unique(calib_probs)=}')
                self.calibrators.append(calib)
                self.n_calibs.append(val_idxs.shape[-1])

        return None

    def predict(self, ds: DictDataset) -> torch.Tensor:
        y_preds = self.alg_interface.predict(ds)
        y_preds_probs = torch.softmax(y_preds, dim=-1)
        y_preds_calib = []
        for i in range(y_preds.shape[0]):
            y_pred_calib = self.calibrators[i].predict_proba(y_preds_probs[i].detach().cpu().numpy())
            y_pred_calib = torch.as_tensor(y_pred_calib, dtype=torch.float32)
            # if y_pred_calib.shape[-1] == 1:
            #     # binary classification
            #     y_pred_calib = torch.cat([1.-y_pred_calib, y_pred_calib], dim=-1)
            #     # y_pred_calib = torch.cat([y_pred_calib, 1.-y_pred_calib], dim=-1)
            if self.config.get('use_calib_offset', True):
                y_pred_calib += 1. / self.n_calibs[i]
            y_pred_calib = torch.log(y_pred_calib)
            y_preds_calib.append(y_pred_calib)
        result = torch.stack(y_preds_calib, dim=0)
        # print(f'{y_preds.shape=}, {result.shape=}')
        return result

    def get_required_resources(self, ds: DictDataset, n_cv: int, n_refit: int, n_splits: int,
                               split_seeds: List[int]) -> RequiredResources:
        return self.alg_interface.get_required_resources(ds, n_cv, n_refit, n_splits, split_seeds)
