import copy
from pathlib import Path
from typing import List, Optional, Dict

import numpy as np
import torch
from scipy.interpolate import interp1d
from sklearn.preprocessing import KBinsDiscretizer

from tab_models import utils
from tab_models.alg_interfaces.alg_interfaces import AlgInterface, SingleSplitAlgInterface
from tab_models.alg_interfaces.base import SplitIdxs, InterfaceResources, RequiredResources
from tab_models.data import DictDataset, TensorInfo
from tab_models.training.logging import Logger


class ClassificationForRegressionAlgInterface(SingleSplitAlgInterface):

    def get_refit_interface(self, n_refit: int, fit_params: Optional[List[Dict]] = None) -> 'AlgInterface':
        classification_alg_interface = self.config['classification_alg_interface']
        classification_refit_alg_interface = classification_alg_interface.get_refit_interface(n_refit,
                                                                                             fit_params=fit_params)
        return ClassificationForRegressionAlgInterface(
            fit_params=fit_params or self.fit_params,
            **utils.update_dict(self.config, dict(classification_alg_interface=classification_refit_alg_interface))
        )

    def fit(self, ds: DictDataset, idxs_list: List[SplitIdxs], interface_resources: InterfaceResources,
            logger: Logger, tmp_folders: List[Optional[Path]], name: str) -> None:
        classification_alg_interface = self.config['classification_alg_interface']
        y_reg = ds.tensors['y']

        assert y_reg.shape[-1] == 1
        assert len(idxs_list) == 1

        n_bins = self.config.get('n_bins', 100)

        # classification_alg_interface can only take a single data set, not multiple ones with different preprocessing
        # therefore, we have to make this single-split and operate on the train+val set
        trainval_idxs = idxs_list[0].train_idxs[0]
        if idxs_list[0].val_idxs is not None:
            trainval_idxs = torch.cat([trainval_idxs, idxs_list[0].val_idxs[0]], dim=0)
        strategy = self.config.get('binning_strategy', 'uniform')
        self.target_binner_ = KBinsDiscretizer(
                n_bins=n_bins,
                strategy=strategy,
                subsample=200_000 if strategy == 'quantile' else None,
                encode="ordinal",
                random_state=idxs_list[0].split_seed,
            )
        self.target_binner_.fit(y_reg[trainval_idxs].numpy())
        y_class = torch.as_tensor(
                        self.target_binner_.transform(y_reg.numpy()),
                        dtype=torch.long
                    )

        ds_class = DictDataset(tensors=utils.update_dict(ds.tensors, {'y': y_class}),
                               tensor_infos=utils.update_dict(ds.tensor_infos,
                                                              {'y': TensorInfo(cat_sizes=[n_bins])}),
                               device=ds.device,
                               n_samples=ds.n_samples)

        classification_alg_interface.fit(ds=ds_class, idxs_list=idxs_list,
                                         interface_resources=interface_resources,
                                         logger=logger, tmp_folders=tmp_folders,
                                         name=name+'_cls-for-reg')

        self.fit_params = classification_alg_interface.get_fit_params()

    def predict(self, ds: DictDataset) -> torch.Tensor:
        classification_alg_interface = self.config['classification_alg_interface']
        y_pred_class = classification_alg_interface.predict(ds)
        # return torch.argmax(y_pred_class, dim=-1).to(torch.float32)
        return torch.as_tensor(np.array([
            self._predict_single_quantile(y_pred_class[i], self.target_binner_, self.config['quantile_alpha'])
            for i in range(y_pred_class.shape[0])]), dtype=torch.float32)[..., None]

    def _predict_single_quantile(self, y_pred, target_binner, quantile_alpha: float):
        # Build the mapper for inverse CDF mapping, from cumulated
        # probabilities to continuous prediction.
        interpolation_knots = self.config.get('interpolation_knots', 'edges')
        interpolation_kind = self.config.get('interpolation_kind', 'linear')
        y_proba = torch.softmax(y_pred, dim=-1).detach().cpu().numpy()
        edges = target_binner.bin_edges_[0]
        n_samples = y_proba.shape[-2]
        if interpolation_knots == "edges":
            y_cdf = np.zeros(shape=(n_samples, edges.shape[0]))
            y_cdf[:, 1:] = np.cumsum(y_proba, axis=1)
            return np.asarray(
                [
                    interp1d(
                        y_cdf_i,
                        edges,
                        kind=interpolation_kind,
                    )(quantile_alpha)
                    for y_cdf_i in y_cdf
                ]
            )
        else:
            midpoints = (edges[1:] + edges[:-1]) / 2
            y_cdf = np.cumsum(y_proba, axis=1)
            return np.asarray(
                [
                    interp1d(
                        y_cdf_i,
                        midpoints,
                        kind=interpolation_kind,
                        bounds_error=False,
                        fill_value=(midpoints[0], midpoints[-1]),
                    )(quantile_alpha)
                    for y_cdf_i in y_cdf
                ]
            )

    def get_required_resources(self, ds: DictDataset, n_cv: int, n_refit: int, n_splits: int, split_seeds: List[int]) -> RequiredResources:
        n_bins = self.config.get('n_bins', 100)
        ds = copy.deepcopy(ds)
        ds.tensor_infos['y'] = TensorInfo(cat_sizes=[n_bins] * ds.tensor_infos['y'].get_n_features())
        return self.config['classification_alg_interface'].get_required_resources(ds, n_cv, n_refit, n_splits, split_seeds)

