from pathlib import Path
from typing import List, Optional, Tuple, Dict

import numpy as np
import torch

from tab_models.alg_interfaces.alg_interfaces import AlgInterface
from tab_models.alg_interfaces.base import SplitIdxs, InterfaceResources, RequiredResources
from tab_models.alg_interfaces.nn_interfaces import NNAlgInterface
from tab_models.data import DictDataset, TaskType
from tab_models.training.logging import Logger
from tab_models.training.metrics import Metrics


class IterativeReinitNNAlgInterface(AlgInterface):

    # todo: refit interface

    def fit(self, ds: DictDataset, idxs_list: List[SplitIdxs], interface_resources: InterfaceResources,
            logger: Logger, tmp_folders: List[Optional[Path]], name: str) \
            -> Optional[List[List[List[Tuple[Dict, float]]]]]:
        reinit_patterns = self.config.get('reinit_patterns', [['plr'], ['plr', 'layer-0']])
        n_reinit_iterations = 1+len(reinit_patterns)
        # n_reinit_iterations = self.config.get(f'n_reinit_iterations', 2)

        n_tt_splits = len(idxs_list)
        n_tv_splits = idxs_list[0].n_trainval_splits

        best_val_errors = [[np.Inf] * n_tv_splits for i in range(n_tt_splits)]
        best_val_iterations = [[0] * n_tv_splits for i in range(n_tt_splits)]

        self.alg_interfaces_ = []

        for i, reinit_pattern in enumerate([[]] + reinit_patterns):
            if i == 0:
                # don't have any importances yet
                alg_interface = NNAlgInterface(**self.config)
            else:
                alg_interface = NNAlgInterface(**self.config,
                                               fixed_init_params=self.alg_interfaces_[-1].model.model.parameters(),
                                               fixed_init_param_patterns=reinit_pattern)

            metrics = Metrics.defaults(y_cat_sizes=ds.tensor_infos['y'].get_cat_sizes())
            val_metric_name = Metrics.default_metric_name(
                task_type=TaskType.CLASSIFICATION if ds.tensor_infos['y'].is_cat() else TaskType.REGRESSION)
            val_metric_name = self.config.get('val_metric_name', val_metric_name)

            # todo: adjust tmp_folders
            eval_results = alg_interface.fit_and_eval(ds, idxs_list, interface_resources, logger, tmp_folders,
                                                      name + f' (importance iteration {i + 1}/{n_reinit_iterations})',
                                                      metrics=metrics, return_preds=False)

            for tt_split_idx in range(n_tt_splits):
                for tv_split_idx in range(n_tv_splits):
                    error = eval_results[tt_split_idx]['metrics', 'val', '1', str(tv_split_idx), val_metric_name]
                    if error < best_val_errors[tt_split_idx][tv_split_idx]:
                        best_val_errors[tt_split_idx][tv_split_idx] = error
                        best_val_iterations[tt_split_idx][tv_split_idx] = i

            self.alg_interfaces_.append(alg_interface)

        self.best_iterations_ = sum(best_val_iterations, [])

        return None

    def predict(self, ds: DictDataset) -> torch.Tensor:
        predictions = [alg_interface.predict(ds) for alg_interface in self.alg_interfaces_]
        n_splits = predictions[0].shape[0]
        return torch.stack([predictions[self.best_iterations_[i]][i] for i in range(n_splits)])

    def get_required_resources(self, ds: DictDataset, n_cv: int, n_refit: int, n_splits: int,
                               split_seeds: List[int]) -> RequiredResources:
        alg_interface = NNAlgInterface(**self.config)
        res = alg_interface.get_required_resources(ds, n_cv, n_refit, n_splits, split_seeds)
        res.time_s *= 2
        return res


class IterativeWeightNNAlgInterface(AlgInterface):

    # todo: refit interface

    def fit(self, ds: DictDataset, idxs_list: List[SplitIdxs], interface_resources: InterfaceResources,
            logger: Logger, tmp_folders: List[Optional[Path]], name: str) \
            -> Optional[List[List[List[Tuple[Dict, float]]]]]:
        n_weight_iterations = 2
        fixed_weights = None

        n_tt_splits = len(idxs_list)
        n_tv_splits = idxs_list[0].n_trainval_splits

        best_val_errors = [[np.Inf] * n_tv_splits for i in range(n_tt_splits)]
        best_val_iterations = [[0] * n_tv_splits for i in range(n_tt_splits)]

        self.alg_interfaces_ = []

        for i in range(n_weight_iterations):
            if i == 0:
                # don't have any importances yet
                alg_interface = NNAlgInterface(**self.config)
            else:
                alg_interface = NNAlgInterface(**self.config, add_fixed_weight_layer=True,
                                               fixed_weights=fixed_weights)

            metrics = Metrics.defaults(y_cat_sizes=ds.tensor_infos['y'].get_cat_sizes())
            val_metric_name = Metrics.default_metric_name(
                task_type=TaskType.CLASSIFICATION if ds.tensor_infos['y'].is_cat() else TaskType.REGRESSION)
            val_metric_name = self.config.get('val_metric_name', val_metric_name)

            # todo: adjust tmp_folders
            eval_results = alg_interface.fit_and_eval(ds, idxs_list, interface_resources, logger, tmp_folders,
                                                      name + f' (importance iteration {i + 1}/{n_weight_iterations})',
                                                      metrics=metrics, return_preds=False)

            for tt_split_idx in range(n_tt_splits):
                for tv_split_idx in range(n_tv_splits):
                    error = eval_results[tt_split_idx]['metrics', 'val', '1', str(tv_split_idx), val_metric_name]
                    if error < best_val_errors[tt_split_idx][tv_split_idx]:
                        best_val_errors[tt_split_idx][tv_split_idx] = error
                        best_val_iterations[tt_split_idx][tv_split_idx] = i

            fixed_weights = alg_interface.get_first_layer_weights(
                with_scale=self.config.get('use_fixed_weight_with_scale', False))
            # print(f'Importances: ', repr(importances.numpy()))

            self.alg_interfaces_.append(alg_interface)

        self.best_iterations_ = sum(best_val_iterations, [])

        return None

    def predict(self, ds: DictDataset) -> torch.Tensor:
        predictions = [alg_interface.predict(ds) for alg_interface in self.alg_interfaces_]
        n_splits = predictions[0].shape[0]
        return torch.stack([predictions[self.best_iterations_[i]][i] for i in range(n_splits)])

    def get_required_resources(self, ds: DictDataset, n_cv: int, n_refit: int, n_splits: int,
                               split_seeds: List[int]) -> RequiredResources:
        alg_interface = NNAlgInterface(**self.config)
        res = alg_interface.get_required_resources(ds, n_cv, n_refit, n_splits, split_seeds)
        res.time_s *= 2
        return res


class IterativeImportanceNNAlgInterface(AlgInterface):

    # todo: refit interface

    def fit(self, ds: DictDataset, idxs_list: List[SplitIdxs], interface_resources: InterfaceResources,
            logger: Logger, tmp_folders: List[Optional[Path]], name: str) \
            -> Optional[List[List[List[Tuple[Dict, float]]]]]:
        n_importance_iterations = self.config.get('n_importance_iterations', 2)
        importances = None

        n_tt_splits = len(idxs_list)
        n_tv_splits = idxs_list[0].n_trainval_splits

        best_val_errors = [[np.Inf] * n_tv_splits for i in range(n_tt_splits)]
        best_val_iterations = [[0] * n_tv_splits for i in range(n_tt_splits)]

        self.alg_interfaces_ = []

        for i in range(n_importance_iterations):
            if i == 0:
                # don't have any importances yet
                alg_interface = NNAlgInterface(**self.config)
            else:
                alg_interface = NNAlgInterface(**self.config, add_importance_layer=True,
                                               feature_importances=importances)

            metrics = Metrics.defaults(y_cat_sizes=ds.tensor_infos['y'].get_cat_sizes())
            val_metric_name = Metrics.default_metric_name(
                task_type=TaskType.CLASSIFICATION if ds.tensor_infos['y'].is_cat() else TaskType.REGRESSION)
            val_metric_name = self.config.get('val_metric_name', val_metric_name)

            # todo: adjust tmp_folders
            eval_results = alg_interface.fit_and_eval(ds, idxs_list, interface_resources, logger, tmp_folders,
                                                      name + f' (importance iteration {i + 1}/{n_importance_iterations})',
                                                      metrics=metrics, return_preds=False)

            for tt_split_idx in range(n_tt_splits):
                for tv_split_idx in range(n_tv_splits):
                    error = eval_results[tt_split_idx]['metrics', 'val', '1', str(tv_split_idx), val_metric_name]
                    if error < best_val_errors[tt_split_idx][tv_split_idx]:
                        best_val_errors[tt_split_idx][tv_split_idx] = error
                        best_val_iterations[tt_split_idx][tv_split_idx] = i

            importances = alg_interface.get_importances()
            # print(f'Importances: ', repr(importances.numpy()))

            self.alg_interfaces_.append(alg_interface)

        self.best_iterations_ = sum(best_val_iterations, [])

        return None

    def predict(self, ds: DictDataset) -> torch.Tensor:
        predictions = [alg_interface.predict(ds) for alg_interface in self.alg_interfaces_]
        n_splits = predictions[0].shape[0]
        return torch.stack([predictions[self.best_iterations_[i]][i] for i in range(n_splits)])

    def get_required_resources(self, ds: DictDataset, n_cv: int, n_refit: int, n_splits: int,
                               split_seeds: List[int]) -> RequiredResources:
        alg_interface = NNAlgInterface(**self.config)
        res = alg_interface.get_required_resources(ds, n_cv, n_refit, n_splits, split_seeds)
        res.time_s *= self.config.get('n_importance_iterations', 2)
        return res
