import random
from pathlib import Path
from typing import Optional, Dict, List, Tuple, Any, Union

import numpy as np
import pandas as pd
import torch

from tab_models import utils
from tab_models.alg_interfaces.base import SubSplitIdxs, InterfaceResources, RequiredResources
from tab_models.data import DictDataset
from tab_models.model.models import PreprocessingFactory
from tab_models.training.logging import Logger
from tab_models.training.metrics import insert_missing_class_columns


class SubSplitLearner:
    def fit(self, ds: DictDataset, sub_split_idxs: SubSplitIdxs, interface_resources: InterfaceResources,
                     logger: Logger, tmp_folder: Optional[Path], name: str, fit_params: Optional[Dict]) \
            -> List[Tuple[Dict, float]]:
        # return List[Tuple[Dict, float]]], i.e., validation scores for every hyperparameter combination
        # (could be number of trees, early stopping epoch, or hyperparameters from hyperparameter optimization)
        # if hyperparams is not None, use these and maybe only return one list element?
        raise NotImplementedError()

    def clone(self) -> 'SubSplitLearner':
        raise NotImplementedError()

    def predict(self, ds: DictDataset, fit_params: Dict) -> torch.Tensor:
        # should return tensor of shape len(ds) x output_shape
        raise NotImplementedError()

    def get_required_resources(self, ds: DictDataset) -> RequiredResources:
        raise NotImplementedError()


class SklearnSubSplitLearner(SubSplitLearner):
    def __init__(self, **config):
        self.config = config
        self.tfm = None
        self.n_classes = None
        self.model = None
        self.train_ds = None

    def fit(self, ds: DictDataset, sub_split_idxs: SubSplitIdxs, interface_resources: InterfaceResources,
                     logger: Logger, tmp_folder: Optional[Path], name: str, fit_params: Optional[Dict]) \
            -> List[Tuple[Dict, float]]:
        # return List[Tuple[Dict, float]]], i.e., validation scores for every hyperparameter combination
        # (could be number of trees, early stopping epoch, or hyperparameters from hyperparameter optimization)
        # if hyperparams is not None, use these and maybe only return one list element?
        torch.manual_seed(sub_split_idxs.alg_seed)  # can be useful for label encoding with randomized permutation
        np.random.seed(sub_split_idxs.alg_seed)
        random.seed(sub_split_idxs.alg_seed)
        print(f'Seeding with seed {sub_split_idxs.alg_seed}')
        print(f'{type(sub_split_idxs.alg_seed)=}')
        self.n_classes = ds.get_n_classes()
        if sub_split_idxs.val_idxs is None:
            trainval_idxs = sub_split_idxs.train_idxs
            # validation indices such that trainval_idxs[rel_val_idxs] is the val_idxs
            # can be used to index trainval_ds later
            rel_val_idxs = torch.zeros(0, dtype=torch.long)
        else:
            trainval_idxs = torch.cat([sub_split_idxs.train_idxs, sub_split_idxs.val_idxs], dim=0)
            rel_val_idxs = torch.arange(sub_split_idxs.train_idxs.shape[0], trainval_idxs.shape[0], dtype=torch.long)

        trainval_ds = ds.get_sub_dataset(trainval_idxs)
        # for filling in missing classes in the train dataset later
        # might not work when the validation set contains classes that the training set doesn't contain
        self.train_ds = ds.get_sub_dataset(sub_split_idxs.train_idxs)

        self.config["tmp_folder"] = tmp_folder

        # create preprocessing factory
        factory = self.config.get('factory', None)
        if factory is None:
            factory = PreprocessingFactory(**self.config)

        # transform according to factory
        fitter = factory.create(ds.tensor_infos)
        self.tfm, trainval_ds = fitter.fit_transform(trainval_ds)

        y = trainval_ds.tensors['y']

        self.model = self._create_sklearn_model(seed=sub_split_idxs.alg_seed,
                                                n_threads=interface_resources.n_threads,
                                                gpu_devices=interface_resources.gpu_devices)
        if self.n_classes == 0 and trainval_ds.tensor_infos['y'].get_n_features() > 1 \
                and self.config.get('use_multioutput_regressor', False):
            from sklearn.multioutput import MultiOutputRegressor
            self.model = MultiOutputRegressor(self.model)  # todo: test this
            y = y.numpy()
        else:
            y = y[:, 0].numpy()

        x_df = trainval_ds.without_labels().to_df()
        cat_col_names = list(x_df.select_dtypes(include='category').columns)
        self._fit_sklearn(x_df=x_df, y=y, val_idxs=rel_val_idxs.numpy(), cat_col_names=cat_col_names)

        if sub_split_idxs.val_idxs is not None:
            return [(dict(), -1.0)]
        else:
            return []

    def _fit_sklearn(self, x_df: pd.DataFrame, y: np.ndarray, val_idxs: np.ndarray,
                     cat_col_names: Optional[List[str]] = None):
        # by default, we ignore the validation set since most sklearn methods do not support it
        n_samples = len(x_df)
        train_mask = np.ones(shape=(n_samples,), dtype=np.bool_)
        train_mask[val_idxs] = False
        x_df = x_df.iloc[train_mask, :]
        y = y[train_mask]
        if cat_col_names is not None and len(cat_col_names) > 0:
            self.model.fit(x_df, y, **{self._get_cat_indexes_arg_name(): cat_col_names})
        else:
            self.model.fit(x_df, y)

    def predict(self, ds: DictDataset, fit_params: Dict) -> torch.Tensor:
        # should return tensor of shape len(ds) x output_shape
        if self.tfm is not None:
            ds = self.tfm.forward_ds(ds)

        x_df = ds.without_labels().to_df()

        if self.n_classes > 0:
            # classification
            y_pred = np.log(self.model.predict_proba(x_df) + 1e-30)
        else:
            # regression
            y_pred = self.model.predict(x_df)
            if len(y_pred.shape) == 1:
                y_pred = y_pred[:, None]

        y_pred = torch.as_tensor(y_pred, dtype=torch.float32)
        # guard against missing classes in the training set
        # (GBDT interfaces don't need this because they get passed n_classes as a parameter)
        y_pred = insert_missing_class_columns(y_pred, self.train_ds)
        return y_pred

    def _create_sklearn_model(self, seed: int, n_threads: int, gpu_devices: List[str]) -> Any:
        # override this in subclasses
        raise NotImplementedError()

    def _get_cat_indexes_arg_name(self) -> str:
        # override this in subclasses if categorical features are supported
        raise NotImplementedError()


class TreeBasedSubSplitLearner(SubSplitLearner):
    def __init__(self, **config):
        self.config = config
        self.tfm = None
        self.n_classes = None
        self.model = None

    def fit(self, ds: DictDataset, sub_split_idxs: SubSplitIdxs, interface_resources: InterfaceResources,
                     logger: Logger, tmp_folder: Optional[Path], name: str, fit_params: Optional[Dict]) \
            -> List[Tuple[Dict, float]]:
        # return List[Tuple[Dict, float]]], i.e., validation scores for every hyperparameter combination
        # (could be number of trees, early stopping epoch, or hyperparameters from hyperparameter optimization)
        # if hyperparams is not None, use these and maybe only return one list element?
        torch.manual_seed(sub_split_idxs.alg_seed)  # can be useful for label encoding with randomized permutation
        np.random.seed(sub_split_idxs.alg_seed)
        self.n_classes = ds.get_n_classes()
        train_ds = ds.get_sub_dataset(sub_split_idxs.train_idxs)
        is_cv = sub_split_idxs.val_idxs is not None
        val_ds = ds.get_sub_dataset(sub_split_idxs.val_idxs) if is_cv else None

        # create preprocessing factory
        factory = self.config.get('factory', None)
        if factory is None:
            factory = PreprocessingFactory(**self.config)

        # transform according to factory
        fitter = factory.create(ds.tensor_infos)
        if is_cv:
            trainval_ds = ds.get_sub_dataset(torch.cat([sub_split_idxs.train_idxs, sub_split_idxs.val_idxs], dim=0))
        else:
            trainval_ds = train_ds
        self.tfm = fitter.fit(trainval_ds)
        train_ds = self.tfm.forward_ds(train_ds)
        if is_cv:
            val_ds = self.tfm.forward_ds(val_ds)

        params = self._get_params()
        if fit_params is not None:
            params = utils.update_dict(params, fit_params)
        gpu_ids = [int(dev_str[len('cuda:'):])
                   for dev_str in interface_resources.gpu_devices if dev_str.startswith('cuda:')]
        if len(gpu_ids) > 0:
            params['gpu_id'] = gpu_ids[0]  # todo: this should at least work for XGBoost
        self.model, val_errors = self._fit(train_ds, val_ds, params=params, seed=sub_split_idxs.alg_seed,
                                           n_threads=interface_resources.n_threads,
                                           val_metric_name=self.config.get('val_metric_name', None),
                                           tmp_folder=tmp_folder)
        if val_errors is None:
            return []
        else:
            return [(dict(n_estimators=i+1), err) for i, err in enumerate(val_errors)]

    def predict(self, ds: DictDataset, fit_params: Dict) -> torch.Tensor:
        # should return tensor of shape len(ds) x output_shape
        if self.tfm is not None:
            ds = self.tfm.forward_ds(ds)
        return self._predict(self.model, ds, self.n_classes, fit_params)

    def _fit(self, train_ds: DictDataset, val_ds: Optional[DictDataset], params: Dict[str, Any], seed: int,
             n_threads: int, val_metric_name: Optional[str] = None,
             tmp_folder: Optional[Path] = None) -> Tuple[Any, Optional[List[float]]]:
        raise NotImplementedError()

    def _predict(self, bst: Any, ds: DictDataset, n_classes: int, other_params: Dict[str, Any]) -> torch.Tensor:
        raise NotImplementedError()

    def _get_params(self) -> Dict[str, Any]:
        raise NotImplementedError()

    # clone() needs to be implemented by subclasses
