from .learner import *
from .metrics import Metrics
from .logging import *
from tab_models.training.coord import HyperparamManager
import numpy as np
import traceback


# todo: could have callback base class such that subclasses get hints at available methods

# callbacks don't have learner as a permanent attribute since the learner has the callbacks as attributes,
# otherwise the learner and the callbacks would only ever get freed when the garbage collector is called

class ParamCheckpointer:
    def __init__(self, n_models, n_parallel):
        self.n_models = n_models
        self.n_parallel = n_parallel
        self.ckpt_params = [None] * self.n_parallel
        self.ckpt_buffers = [None] * self.n_parallel

    def save(self, parallel_idx: int, model: Layer):
        start = self.n_models * parallel_idx
        end = self.n_models * (parallel_idx + 1)
        with torch.no_grad():
            for ckpt, values in [(self.ckpt_params, model.parameters()), (self.ckpt_buffers, model.buffers())]:
                if ckpt[parallel_idx] is None:
                    ckpt[parallel_idx] = [v[start:end].clone() for v in values]
                else:
                    for c, v in zip(ckpt[parallel_idx], values):
                        c.copy_(v[start:end])

    def restore(self, parallel_idx: int, model: Layer):
        start = self.n_models * parallel_idx
        end = self.n_models * (parallel_idx + 1)
        with torch.no_grad():
            for ckpt, values in [(self.ckpt_params, model.parameters()), (self.ckpt_buffers, model.buffers())]:
                if ckpt[parallel_idx] is not None:
                    for c, v in zip(ckpt[parallel_idx], values):
                        # print(f'Restore diff: {v[start:end]-c}')
                        v[start:end] = c

    def save_all(self, model: Layer):
        for i in range(self.n_parallel):
            self.save(i, model)

    def restore_all(self, model: Layer):
        for i in range(self.n_parallel):
            self.restore(i, model)


class StopAtEpochsCallback:
    name = 'stop_at_epochs'

    def __init__(self, stop_epochs, n_models, logger: Logger):
        self.stop_epochs = stop_epochs
        self.final_stop_epoch = np.max(stop_epochs)
        self.ckpt = ParamCheckpointer(n_models=n_models, n_parallel=len(stop_epochs))
        self.logger = logger

    def before_fit(self, learner: Learner):
        for i, ep in enumerate(self.stop_epochs):
            if ep == 0:
                self.ckpt.save(i, learner.model)

    def after_epoch(self, learner: Learner):
        self.logger.log(2, f'Refit Epoch {learner.progress.epoch + 1}/{self.final_stop_epoch}')
        epoch = learner.progress.epoch + 1
        if epoch == self.final_stop_epoch:
            self.ckpt.restore_all(learner.model)
            raise CancelTrainException()

        for i, ep in enumerate(self.stop_epochs):
            if ep == epoch:
                self.ckpt.save(i, learner.model)


# todo: could split loss computation + logging, best epoch tracking + checkpointing into separate callbacks
# then can also add an option to compute and log train loss
class ValidationCallback:
    name = 'val'

    def __init__(self, ds: DictDataset, val_idxs: torch.Tensor, metric_name: str, logger: Logger,
                 n_models: int, n_parallel: int, save_best_params=True, val_batch_size=256):
        # todo: option to log only every few seconds at lower log-level?
        ds_x, ds_y = ds.split_xy()
        self.val_x_dl = ParallelDictDataLoader(ds_x, val_idxs, batch_size=val_batch_size)
        self.val_idxs = val_idxs
        self.val_y = ds_y.get_batch(val_idxs)['y']
        self.save_best_params = save_best_params
        self.metric_name = metric_name
        self.logger = logger
        self.n_models = n_models
        self.n_parallel = n_parallel
        # mean val errors will not be accurate if all epochs after this yield NaN
        self.best_mean_val_errors = [np.Inf] * n_parallel
        # epoch 0 counts as before training, epoch 1 is first epoch
        self.best_val_epochs = [0] * n_parallel

        # list[tensor], one tensor for each epoch, tensor contains one scalar for each model in ensemble
        self.val_errors_list = []
        self.train_losses_list = []

        self.ckpt = ParamCheckpointer(n_models=n_models, n_parallel=n_parallel)

    def before_epoch(self, learner: Learner):
        self.train_losses_list.append([])

    def after_loss(self, learner: Learner):
        self.train_losses_list[-1].append(learner.loss.item())

    def restore_best_params(self, learner: Learner):
        if not self.save_best_params:
            raise RuntimeError('ValidationCallback: Cannot restore best params when using save_best_params=False')

        self.ckpt.restore_all(learner.model)

    def before_fit(self, learner: Learner):
        if self.save_best_params:
            self.ckpt.save_all(learner.model)

    def after_epoch(self, learner: Learner):
        y_pred = learner.predict_dl(self.val_x_dl)
        val_errors = torch.as_tensor([Metrics.apply(y_pred[i, :, :], self.val_y[i, :, :], self.metric_name)
                                      for i in range(y_pred.shape[0])])
        self.val_errors_list.append(val_errors)
        mean_val_errors = val_errors.view(self.n_parallel, self.n_models).mean(dim=-1)
        mean_val_error = mean_val_errors.mean().item()
        self.logger.log(2,
            f'Epoch {learner.progress.epoch + 1}/{learner.progress.max_epochs}: val error = {mean_val_error:6.4f}')

        for i in range(self.n_parallel):
            if mean_val_errors[i] <= self.best_mean_val_errors[i]:
                # use <= on purpose such that latest epoch among tied best epochs is kept
                # this has been slightly beneficial for accuracy in previous experiments
                self.best_mean_val_errors[i] = mean_val_errors[i]
                self.best_val_epochs[i] = learner.progress.epoch + 1  # todo: ensure that there is no off-by-1 error
                if self.save_best_params:
                    self.ckpt.save(i, learner.model)


class HyperparamCallback:
    name = 'hyper'

    def __init__(self, hp_manager: HyperparamManager):
        self.hp_manager = hp_manager

    def before_pred(self, learner):
        # print(list(learner.model.parameters())[-1])
        # print(list(learner.model.parameters())[-1][0, -1].item())
        self.hp_manager.update_hypers(learner)

    def after_loss(self, learner):
        reg_terms = self.hp_manager.reg_terms
        if len(reg_terms) > 0:
            learner.loss = learner.loss + sum(reg_terms)


class SharpnessAwareMinimizationCallback:
    # following https://github.com/davda54/sam
    name = 'sam'

    def __init__(self, opt: OptimizerBaseOld, hp_manager: HyperparamManager, sam_adaptive=False, **config):
        self.opt = opt
        self.hp_manager = hp_manager
        self.sam_adaptive = sam_adaptive
        # currently don't use separate scopes for separate parameters
        self.sam_rho_getters = [hp_manager.register_hyper(name='sam_rho', scope=group['params'][0].context.scope)
                                for group in opt.param_groups]
        self.e_w_state = {p: None for group in self.opt.param_groups for p in group['params']}

    def after_backward(self, learner: Learner):
        with torch.no_grad():
            grad_norm = torch.norm(
                torch.stack([
                    ((torch.abs(p) if self.sam_adaptive else 1.0) * p.grad).norm(p=2)
                    for group in self.opt.param_groups for p in group["params"]
                    if p.grad is not None
                ]),
                p=2
            )
            for group, rho_getter in zip(self.opt.param_groups, self.sam_rho_getters):
                scale = rho_getter() / (grad_norm + 1e-12)

                for p in group["params"]:
                    if p.grad is None:
                        continue
                    e_w = (torch.pow(p, 2) if self.sam_adaptive else 1.0) * p.grad * scale
                    p.add_(e_w)  # climb to the local maximum "w + e(w)"
                    self.e_w_state[p] = e_w

            self.opt.zero_grad()

        output = learner.model(learner.batch)
        loss = learner.criterion(output['x_cont'], output['y']).sum()
        loss.backward()

        with torch.no_grad():
            for group in self.opt.param_groups:
                for p in group["params"]:
                    if p.grad is None: continue
                    p.sub_(self.e_w_state[p])


class L1L2RegCallback:
    name = 'l1l2reg'

    def __init__(self, hp_manager: HyperparamManager, model: Layer):
        self.hp_manager = hp_manager
        self.params = list(model.parameters())
        self.l1_getters = [self.hp_manager.register_hyper('l1_reg', p.context.scope, default=0.0)
                           for p in self.params]
        self.l2_getters = [self.hp_manager.register_hyper('l2_reg', p.context.scope, default=0.0)
                           for p in self.params]

    def after_backward(self, learner):
        # params = list(learner.model.parameters())
        # if self.l1_getters is None or self.l2_getters is None:
        #     self.l1_getters = [self.hp_manager.register_hyper('l1_reg', p.context.scope, default=0.0)
        #                        for p in params]
        #     self.l2_getters = [self.hp_manager.register_hyper('l2_reg', p.context.scope, default=0.0)
        #                        for p in params]

        for l1_getter, l2_getter, p in zip(self.l1_getters, self.l2_getters, self.params):
            l1_reg = l1_getter() * p.hyper_factors.get('l1_reg', 1.0)
            l2_reg = l2_getter() * p.hyper_factors.get('l2_reg', 1.0)

            with torch.no_grad():
                if l1_reg != 0.0:
                    p.grad += l1_reg * torch.sgn(p)
                if l2_reg != 0.0:
                    p.grad += (2.0 * l2_reg) * p
        self.hp_manager.update_hypers(learner)


class TemporaryCallbacks:
    """ A context manager to introduce callbacks to a Learner and then restore the previous callbacks """
    def __init__(self, learner, callbacks):
        self.learner = learner
        self.callbacks = callbacks
        self.old_callbacks = []

    def __enter__(self):
        self.old_callbacks = [self.learner.get_callback(cb.name) for cb in self.callbacks]
        for cb in self.callbacks:
            self.learner.set_callback(cb.name, cb)

    def __exit__(self, *args):
        for cb, old_cb in zip(self.callbacks, self.old_callbacks):
            self.learner.set_callback(cb.name, old_cb)


