import copy
from tab_models.data.data import *
from pathlib import Path
from tab_models.model.base import Layer
from tab_models.optim.optimizers import OptimizerBaseOld
from .scheduling import LearnerProgress


class CancelBatchException(Exception):
    pass


class CancelEpochException(Exception):
    pass


class CancelTrainException(Exception):
    pass


# todo: theoretically, could make progress also managed from a callback
# todo: have method Learner.to()  (would also be useful for saving/loading)
# todo: improve saving and loading, also allow to save the learner's state itself (when interrupting training)


# for different training loop, the Learner can be inherited from
class Learner:
    # Learner is inspired by the fastai Learner
    def __init__(self, model: Layer, criterion, data_loader: ParallelDictDataLoader, opt: OptimizerBaseOld,
                 callbacks=None):
        self.model = model
        self.opt = opt
        self.data_loader = data_loader
        self.criterion = criterion
        self.callbacks = []
        self.batch = None
        self.output = None
        self.loss = None
        self.progress = LearnerProgress()
        self.dataset_size = self.data_loader.n_samples
        self.device = data_loader.ds.device  # todo

        self.add_callbacks(callbacks)

    def add_callback(self, callback):
        self.add_callbacks([callback])

    def add_callbacks(self, callbacks):
        self.callbacks += callbacks
        self.register_callbacks()
        self.sort_callbacks()

    def sort_callbacks(self):
        sorted_callbacks = []
        remaining_callbacks = copy.copy(self.callbacks)
        # use simple O(n^2) topological sorting algorithm (not maximally efficient, but we have few callbacks anyway)
        # the algorithm repeatedly passes through the list and tries to find an element without predecessor
        while len(remaining_callbacks) > 0:
            cb_names = [cb.name for cb in remaining_callbacks]
            for i, cb in enumerate(remaining_callbacks):
                has_predecessor = any([ra in cb_names for ra in getattr(cb, 'run_after', [])]) or \
                                  any([cb.name in getattr(other_cb, 'run_before', []) for other_cb in
                                       remaining_callbacks])
                if not has_predecessor:
                    sorted_callbacks.append(cb)
                    remaining_callbacks.pop(i)
                    break
            else:  # if break was never called in loop
                raise RuntimeError('Callbacks in Learner could not be sorted')

        self.callbacks = sorted_callbacks

    def register_callbacks(self):
        for cb in self.callbacks:
            setattr(self, 'cb_' + cb.name, cb)

    def get_callback(self, cb_name):
        return getattr(self, 'cb_' + cb_name, None)

    def set_callback(self, cb_name, cb):
        assert cb is None or cb_name == cb.name
        idxs = [i for i, cb in enumerate(self.callbacks) if cb.name == cb_name]
        if len(idxs) > 0:
            # there should only be one such callback, but just in case...
            for idx in reversed(idxs):  # pop them from back to front, otherwise the indexes change...
                self.callbacks.pop(idx)
            delattr(self, 'cb_' + cb_name)
        if cb is not None:
            self.add_callback(cb)

    def callback(self, callback_name):
        for cb in self.callbacks:
            method = getattr(cb, callback_name, None)
            if callable(method):
                method(self)

    def train_one_batch(self, batch):
        self.batch = batch
        try:
            # todo: do we need a self.batch_orig?
            self.callback('before_pred')
            self.output = self.model(self.batch)
            self.callback('after_pred')
            self.loss = self.criterion(self.output['x_cont'], self.output['y']).sum()  # sum over parallel dimension
            self.callback('after_loss')
            self.loss.backward()
            self.callback('after_backward')
            self.opt.step()
            self.callback('after_step')
            self.opt.zero_grad()

        except CancelBatchException:
            pass

    def train_one_epoch(self):
        try:
            for i, batch in enumerate(self.data_loader):
                self.callback('before_batch')

                self.train_one_batch(batch)

                self.progress.epoch_samples += self.batch['y'].shape[-2]
                self.progress.total_samples += self.batch['y'].shape[-2]
                self.progress.epoch_float = self.progress.total_samples \
                                            / self.data_loader.get_num_iterated_samples()
                self.progress.epoch_steps += 1
                self.progress.total_steps += 1

                self.callback('after_batch')

                # "delete" self.output, self.loss, self.batch such that their memory can be released
                self.output = self.loss = self.batch = None

        except CancelEpochException:
            pass

    def fit(self, max_epochs):
        """allow n_epochs = None, then training must be terminated by an exception from a callback"""
        self.progress = LearnerProgress()
        self.progress.max_epochs = max_epochs

        # todo: move everything from here into a continue_fit() method?
        #  (in order to continue training with a serialized learner)
        try:
            self.model.train()
            self.callback('before_fit')
            while self.progress.max_epochs is None or self.progress.epoch < self.progress.max_epochs:
                self.callback('before_epoch')
                self.train_one_epoch()
                self.callback('after_epoch')
                self.progress.epoch += 1
                self.progress.epoch_samples = 0
                self.progress.epoch_steps = 0
                # todo: in order to save training after an epoch, an update should be performed here
        except CancelTrainException:
            pass
        finally:
            self.callback('after_fit')

    # def predict(self, tensors, max_bs=100000):
    #     return self.predict_dl(TaskDataLoader(DictDataset(tensors, tensor_infos=None), batch_size=max_bs, shuffle=False))
    #
    # def predict_batch(self, X):
    #     old_training = self.model.training
    #     self.model.eval()
    #     with torch.no_grad():
    #         result = self.model(X)
    #         self.model.train(old_training)
    #         return result

    def predict_dl(self, dl):
        old_training = self.model.training
        self.model.eval()
        with torch.no_grad():
            results = []
            for xb in dl:
                results.append(self.model(xb)['x_cont'])
            self.model.train(old_training)
            return torch.cat(results, dim=-2)

    def save(self, folder):
        torch.save(self.model.state_dict(), Path(folder) / 'model_state.p')
        self.opt.save(Path(folder) / 'opt_state.p')

    def load(self, folder):
        self.model.load(torch.load(Path(folder) / 'model.p'))
        self.opt.load(Path(folder) / 'opt_state.p')

    def get_state(self):
        return {'model': copy.deepcopy(self.model.state_dict()), 'opt': copy.deepcopy(self.opt.state_dict())}

    def set_state(self, state):
        self.model.load_state_dict(state['model'])
        self.opt.load_state_dict(state['opt'])  # todo: need to copy here?

    # from https://discuss.pytorch.org/t/how-do-i-check-the-number-of-parameters-of-a-model/4325/7
    def count_parameters(self):
        # todo: does not work properly if models are vectorized
        return sum(p.numel() for p in self.model.parameters() if p.requires_grad)



