from tab_models.model.categorical import *
from tab_bench.run.results import *
from tab_models.training.metrics import *
from tab_models.training.logging import Logger

from tab_bench.scheduling.resources import NodeResources
from tab_models.alg_interfaces.base import RequiredResources


class AlgWrapper:  # or AlgWrapper
    def __init__(self, resource_computation=None, **config):
        self.config = config
        self.resource_computation = resource_computation

    def run(self, task_package: TaskPackage, logger: Logger, assigned_resources: NodeResources,
            tmp_folders: List[Path]) -> List[ResultManager]:
        raise NotImplementedError()

    def get_required_resources(self, task_package: TaskPackage) -> RequiredResources:
        if self.resource_computation is not None:
            return self.resource_computation.get_required_resources(task_package)
        raise NotImplementedError()

    def set_resource_computation(self, resource_computation):
        self.resource_computation = resource_computation

    def get_max_n_vectorized(self, task_info: TaskInfo) -> int:
        return 1  # override if AlgWrapper can exploit vectorization

    def add_defaults(self, **kwargs):
        self.config = utils.update_dict(kwargs, self.config)


class SingleSplitWrapper(AlgWrapper):
    def __init__(self, factory=IdentityFactory(), **config):
        super().__init__(**config)
        self.factory = factory

    def run_on_split(self, task_package: TaskPackage, split_info: SplitInfo, trainval: DictDataset, test: DictDataset,
                     rm: ResultManager, logger: Logger, assigned_resources: NodeResources):
        raise NotImplementedError()  # needs to be overridden in subclasses

    def run(self, task_package: TaskPackage, rms: List[ResultManager], logger: Logger,
            assigned_resources: NodeResources):
        task = task_package.task_info.load_task(task_package.paths)
        fitter = self.factory.create(task.ds.tensor_infos)
        for rm, split_info in zip(rms, task_package.split_infos):
            # this will usually be called with len(task_package.split_infos) == 1, but do a loop for safety
            test_split = split_info.splitter.split_ds(task.ds)
            trainval, test = test_split.get_sub_ds(0), test_split.get_sub_ds(1)

            tfm, trainval = fitter.fit_transform(trainval)
            test = tfm.forward_ds(test)

            logger.log(1, f'Running on split {split_info.id} of task {task_package.task_info.task_desc}')
            self.run_on_split(task_package, split_info, trainval, test, rm, logger, assigned_resources)


# want to have:
# - more general / easy ResourceComputation
# - generic thread-allocation parameters for such a ResourceComputation
#   that allow to allocate more threads for larger workloads
# - better NodeResources class that supports mps or perhaps a new class that summarizes the allocated resources
# - should the resource estimation be moved to AlgInterface?
#   Then, we would need to instantiate an AlgInterface in the wrapper to do the estimation
# - maybe a code that estimates RAM (and time) constants? With fake data sets?

# better ResourceComputation:
# have identical components for CPU and GPU, and maybe also for RAM and time
# components:
# - dataset size
# - factory (model) size
# - RAM for forward (and backward) pass
# - generic calculation (constant, per-tree, per-class, per-sample),
# for the NN we might also need to include the batch size, number of epochs, etc.
# what about the number of threads etc.?
# want to have one per device?

# better NodeResources:
# maybe just have a dict with the devices that are being referred to by the array?


class TreeResourceComputation:
    def __init__(self, time_factor: float, factory: FitterFactory, n_estimators: int, multiclass_multitree: bool,
                 supports_cv_parallelization: bool, parallelizable_fraction: float = 1.0,
                 max_n_threads=8.0, max_split_threads=1.0,  ram_ds_factor: float = 3.0,
                 ram_offset_gb: float = 0.3, tree_size_gb: float = 0.002, tree_size_gb_per_sample: float = 0.0,
                 use_gpu: bool = False):
        self.time_factor = time_factor
        self.factory = factory
        self.n_estimators = n_estimators
        self.multiclass_multitree = multiclass_multitree
        self.supports_cv_parallelization = supports_cv_parallelization,
        self.parallelizable_fraction = parallelizable_fraction
        self.max_n_threads = max_n_threads
        self.max_split_threads = max_split_threads
        self.ram_ds_factor = ram_ds_factor
        self.ram_offset_gb = ram_offset_gb
        self.tree_size_gb = tree_size_gb
        self.tree_size_gb_per_sample = tree_size_gb_per_sample
        self.use_gpu = use_gpu

    def get_required_resources(self, task_package: TaskPackage) -> RequiredResources:
        # in hyperopt method also on number of steps
        # moreover it should depend on n_threads, and scaling law should be able to be configured
        # should allow n_threads to depend on the task_info  (based on certain thresholds and possibly scaling law)
        # include a time_factor depending on the method
        n_splits = len(task_package.split_infos)
        n_samples = task_package.task_info.n_samples
        ds_size_gb = task_package.task_info.get_ds_size_gb()
        n_classes = task_package.task_info.tensor_infos['y'].get_cat_sizes()[0].item()
        tensor_infos = task_package.task_info.tensor_infos
        fitter, out_tensor_infos = self.factory.create_transform(tensor_infos)
        n_features = sum([ti.get_n_features() for key, ti in out_tensor_infos.items()
                          if key in ['x_cont', 'x_cat']])

        n_tree_repeats = 1 if n_classes <= 2 or not self.multiclass_multitree else n_classes
        n_trees = self.n_estimators * n_tree_repeats

        n_cv = task_package.n_cv
        n_refit = task_package.n_refit

        # THREADS
        # todo: allow additional factor, maybe even based on RAM usage or predicted time?
        thread_factor = max(n_cv, n_refit)
        if thread_factor > self.max_n_threads or not self.supports_cv_parallelization:
            thread_factor = 1.0
        n_threads = min(self.max_n_threads, thread_factor * self.max_split_threads)

        # TIME
        # could depend on size of dataset, but we ignore that for now
        thread_time_factor = (1.0 - self.parallelizable_fraction) / thread_factor \
                             + self.parallelizable_fraction / n_threads
        # incorporate number of hyperopt iterations via time_factor
        time_approx = self.time_factor * thread_time_factor * n_samples * n_trees * n_features \
                      * (n_cv + n_refit) * n_splits

        # RAM
        thread_ram_factor = thread_factor
        cpu_ram_gb = self.ram_offset_gb + self.ram_ds_factor * ds_size_gb + n_trees * self.tree_size_gb \
                        + n_trees * n_samples * self.tree_size_gb_per_sample
        cpu_ram_gb *= thread_ram_factor

        if self.use_gpu:
            # don't reserve gpu ram explicitly since gpu_usage=1.0 reserves a whole GPU
            return RequiredResources(time_s=time_approx, n_threads=n_threads, cpu_ram_gb=0.5, n_gpus=1,
                                     gpu_usage=1.0, gpu_ram_gb=0.0)
        else:
            return RequiredResources(time_s=time_approx, n_threads=n_threads, cpu_ram_gb=cpu_ram_gb)


class SimpleCPUResourceComputation:
    def __init__(self, time_factor: float, supports_cv_parallelization: bool = False,
                 parallelizable_fraction: float = 1.0, max_n_threads=1.0, max_split_threads=1.0,
                 ram_ds_factor: float = 3.0, ram_offset_gb: float = 0.3):
        self.time_factor = time_factor
        self.supports_cv_parallelization = supports_cv_parallelization,
        self.parallelizable_fraction = parallelizable_fraction
        self.max_n_threads = max_n_threads
        self.max_split_threads = max_split_threads
        self.ram_ds_factor = ram_ds_factor
        self.ram_offset_gb = ram_offset_gb

    def get_required_resources(self, task_package: TaskPackage) -> RequiredResources:
        # in hyperopt method also on number of steps
        # moreover it should depend on n_threads, and scaling law should be able to be configured
        # should allow n_threads to depend on the task_info  (based on certain thresholds and possibly scaling law)
        # include a time_factor depending on the method
        n_splits = len(task_package.split_infos)
        n_samples = task_package.task_info.n_samples
        ds_size_gb = task_package.task_info.get_ds_size_gb()

        n_cv = task_package.n_cv
        n_refit = task_package.n_refit

        # THREADS
        thread_factor = max(n_cv, n_refit)
        if thread_factor > self.max_n_threads or not self.supports_cv_parallelization:
            thread_factor = 1.0
        n_threads = min(self.max_n_threads, thread_factor * self.max_split_threads)

        # TIME
        # could depend on size of dataset, but we ignore that for now
        thread_time_factor = (1.0 - self.parallelizable_fraction) / thread_factor \
                             + self.parallelizable_fraction / n_threads
        # incorporate number of hyperopt iterations via time_factor
        time_approx = self.time_factor * thread_time_factor * n_samples \
                      * (n_cv + n_refit) * n_splits

        # RAM
        thread_ram_factor = thread_factor
        cpu_ram_gb = self.ram_offset_gb + self.ram_ds_factor * ds_size_gb
        cpu_ram_gb *= thread_ram_factor

        return RequiredResources(time_s=time_approx, n_threads=n_threads, cpu_ram_gb=cpu_ram_gb)










